#!/bin/sh

### Brief install instructions for convenience ###

# Partition disk:
# cfdisk /dev/nvme0n1
#
# Format paritions:
# example:
#  mkfs.ext4 -L ROOT /dev/sda2
# mkfs.ext4 -L HOME /dev/sda3
# mkfs.ext4 -L BOOT /dev/sda1
# mkswap -L SWAP /dev/sda4
#
# mkfs.fat -F 32 /dev/sda1
# fatlabel /dev/sda1 ESP
# swapon /dev/disk/by-label/SWAP
# mount /dev/disk/by-label/ROOT /mnt
# mkdir /mnt/boot
# mkdir /mnt/home
# mount /dev/disk/by-label/HOME /mnt/home
# mount /dev/disk/by-label/BOOT /mnt/boot
# mkdir /mnt/boot/efi
# mount /dev/disk/by-label/ESP /mnt/boot/efi
#
# Connect through wifi:
# iwctl
# ping artixlinux.org
#
# Update system clock:
# rc-service ntpd start
#
# Install the base system and kernel:
# basestrap /mnt base base-devel openrc elogind-openrc
# basestrap /mnt linux linux-firmware
#
# Generate file system table:
# fstabgen -U /mnt >> /mnt/etc/fstab
# # (verify the fstab file)
#
# Enter new system:
# artix-chroot /mnt
#
# Configure system clock:
# ln -sf /usr/share/zoneinfo/Europe/Belgrade /etc/localtime
# hwclock --systohc
#
# Enable locales:
# pacman -S vim
# vim /etc/locale.gen
# locale-gen
#
# Install boot loader:
# pacman -S grub os-prober efibootmgr
# grub-install --target=x86_64-efi --efi-directory=/boot/efi --bootloader-id=grub
# grub-mkconfig -o /boot/grub/grub.cfg
#
# Add users:
# (set root passwd)
# passwd
#
# useradd -m kappa
# passwd kappa
#
# Network config: (maybe this should be changed for network manager)
# vim /etc/hostname # add a hostname
# vim /etc/hosts
# # 127.0.0.1        localhost
# # ::1              localhost
# # 127.0.1.1        myhostname.localdomain  myhostname
# vim /etc/conf.d/hostname
# hostname="myhostname"
#
# pacman -S dhcpcd iwd dhcpcd-openrc iwd-openrc
# rc-update add dhcpcd
# rc-update add iwd
#
# Reboot system:
# exit
# umount -R /mnt
# reboot


# Luke's Auto Rice Bootstrapping Script (LARBS)
# by Luke Smith <luke@lukesmith.xyz>
# License: GNU GPLv3

### OPTIONS AND VARIABLES ###

progsfile="https://git.bonsai.cool/kayprish/bs/raw/branch/master/progs.csv"
aurhelper="yay"
export TERM=ansi

### FUNCTIONS ###

error() {
	# Log to stderr and exit with failure.
	printf "%s\n" "$1" >&2
	exit 1
}

welcomemsg() {
	whiptail --title "Welcome!" \
		--msgbox "Welcome to Luke's Auto-Rice Bootstrapping Script!\\n\\nThis script will automatically install a fully-featured Linux desktop, which I use as my main machine.\\n\\n-Luke" 10 60

	whiptail --title "Important Note!" --yes-button "All ready!" \
		--no-button "Return..." \
		--yesno "Be sure the computer you are using has current pacman updates and refreshed Arch keyrings.\\n\\nIf it does not, the installation of some programs might fail." 8 70
}

getuserandpass() {
	# Prompts user for new username and password.
	name=$(whiptail --inputbox "First, please enter a name for the user account." 10 60 3>&1 1>&2 2>&3 3>&1) || exit 1
	while ! echo "$name" | grep -q "^[a-z_][a-z0-9_-]*$"; do
		name=$(whiptail --nocancel --inputbox "Username not valid. Give a username beginning with a letter, with only lowercase letters, - or _." 10 60 3>&1 1>&2 2>&3 3>&1)
	done
	pass1=$(whiptail --nocancel --passwordbox "Enter a password for that user." 10 60 3>&1 1>&2 2>&3 3>&1)
	pass2=$(whiptail --nocancel --passwordbox "Retype password." 10 60 3>&1 1>&2 2>&3 3>&1)
	while ! [ "$pass1" = "$pass2" ]; do
		unset pass2
		pass1=$(whiptail --nocancel --passwordbox "Passwords do not match.\\n\\nEnter password again." 10 60 3>&1 1>&2 2>&3 3>&1)
		pass2=$(whiptail --nocancel --passwordbox "Retype password." 10 60 3>&1 1>&2 2>&3 3>&1)
	done
}

usercheck() {
	! { id -u "$name" >/dev/null 2>&1; } ||
		whiptail --title "WARNING" --yes-button "CONTINUE" \
			--no-button "No wait..." \
			--yesno "The user \`$name\` already exists on this system. LARBS can install for a user already existing.\\n\\nLARBS will change $name's password to the one you just gave." 14 70
}

selectcategories() {
	([ -f "$progsfile" ] && cp "$progsfile" /tmp/progs.csv) ||
		curl -Ls "$progsfile" | sed '/^#/d' >/tmp/progs.csv
	selectedcats=$(whiptail --title "Select package categories" --separate-output \
		--checklist "Categories" 20 100 12 \
		"amd" "AMD specific packages" ON \
		"anki" "Anki" ON \
		"backup" "Packages for backing the system up (borg)" ON \
		"crypto" "Cryptocurrencies" ON \
		"cysec" "Packages i use for my work in Cybersecurity" ON \
		"dumb" "silly things I like to have" ON \
		"editing" "Audio/video editing" ON \
		"electronics" "Packages used for electronics and FPGA work" ON \
		"fonts" "Fonts" ON \
		"go" "Packages for work with golang" ON \
		"gui" "Packages needed for my GUI and dotfiles to work" ON \
		"haskell" "Haskell and things needed for it" ON \
		"latex" "LaTeX" ON \
		"lisp" "All sorts of Lisps" ON \
		"lowlev" "Packages for programming in low-level languages (C,C++,Rust,etc)" ON \
		"mail" "Mail client and related programs" ON \
		"matrix" "Matrix clients" ON \
		"ml" "Machine learning" ON \
		"office" "Office packages" ON \
		"passwords" "Password manager and related" ON \
		"python" "Packages for python programming" ON \
		"retro" "Retrocomputing" ON \
		"rss" "RSS" ON \
		"stats" "Statistics (R)" ON \
		"system" "Packages I expect on my system" ON \
		"tcl" "Tcl/Tk and related" ON \
		"torrent" "Torrent clients" ON \
		"virtual" "Virtual machine packages" ON \
		"webdev" "Web-development" ON \
		"www" "Browsers" ON 3>&1 1>&2 2>&3 3>&1) || exit 1
}

preinstallmsg() {
	whiptail --title "Let's get this party started!" --yes-button "Let's go!" \
		--no-button "No, nevermind!" \
		--yesno "The rest of the installation will now be totally automated, so you can sit back and relax.\\n\\nIt will take some time, but when done, you can relax even more with your complete system.\\n\\nNow just press <Let's go!> and the system will begin installation!" 13 60 || {
		clear
		exit 1
	}
}

adduserandpass() {
	# Adds user `$name` with password $pass1.
	whiptail --infobox "Adding user \"$name\"..." 7 50
	useradd -m -g wheel -s /bin/bash "$name" >/dev/null 2>&1 ||
		usermod -a -G wheel "$name" && mkdir -p /home/"$name" && chown "$name":wheel /home/"$name"
	export repodir="/home/$name/.local/src"
	mkdir -p "$repodir"
	chown -R "$name":wheel "$(dirname "$repodir")"
	echo "$name:$pass1" | chpasswd
	unset pass1 pass2
}

refreshkeys() {
	case "$(readlink -f /sbin/init)" in
	*systemd*) # May end up removing the Arch option
		whiptail --infobox "Refreshing Arch Keyring..." 7 40
		pacman --noconfirm -S archlinux-keyring
		;;
	*)
		whiptail --infobox "Enabling Arch and Artix Omniverse Repositories for more a more extensive software collection..." 7 40
		pacman --noconfirm --needed -S \
			artix-keyring artix-archlinux-support
		grep -q "^\[extra\]" /etc/pacman.conf ||
			echo "[extra]
Include = /etc/pacman.d/mirrorlist-arch" >>/etc/pacman.conf
		grep -q "^\[extra\]" /etc/pacman.conf ||
			echo "[omniverse]
		]
Server = https://artix.sakamoto.pl/omniverse/$arch
Server = https://eu-mirror.artixlinux.org/omniverse/$arch
Server = https://omniverse.artixlinux.org/$arch" >>/etc/pacman.conf
		pacman -Sy --noconfirm
		pacman-key --populate archlinux
		;;
	esac
}

manualinstall() {
	# Installs $1 manually. Used only for AUR helper here.
	# Should be run after repodir is created and var is set.
	pacman -Qq "$1" && return 0
	whiptail --infobox "Installing \"$1\" manually." 7 50
	sudo -u "$name" mkdir -p "$repodir/$1"
	sudo -u "$name" git -C "$repodir" clone --depth 1 --single-branch \
		--no-tags -q "https://aur.archlinux.org/$1.git" "$repodir/$1" ||
		{
			cd "$repodir/$1" || return 1
			sudo -u "$name" git pull --force origin master
		}
	cd "$repodir/$1" || exit 1
	sudo -u "$name" makepkg --noconfirm -si || return 1
}

maininstallations() {
	aurinstalled=$(pacman -Qqm)
	# Field 1: install-type, Field 2: program, Field 3: comment
	# Install pacman packages:
	pacman --noconfirm -Rnd vim # Remove vim as it will conflict with gvim
	awk -F',' -v tags="${selectedcats}" '
	BEGIN {
	    # Convert selectedcats string to an array in awk
	    n = split(tags, sel_cats, " ");
	    for (i = 1; i <= n; i++) {
	        selected[sel_cats[i]] = 1;
	    }
	}
	{
	    # Assuming tags are in the 3rd column
	    split($3, tag_list, "-");
	    for (i in tag_list) {
	        if (tag_list[i] in selected) {
	            print $0;
	            next;
	        }
	    }
	}' /tmp/progs.csv >/tmp/filtered-progs.csv

	awk -F "," '$1 ~ /^$/ { print $2 }'  </tmp/filtered-progs.csv | xargs pacman --noconfirm --needed -S
	awk -F "," '$1 ~ /^A$/ { print $2 }' </tmp/filtered-progs.csv | xargs sudo -u "$name" "$aurhelper" --noconfirm --needed -S
}

finalize() {
	echo "The bootstrap script is complete, have fun with your new system!\\n\\nYou can log in as your user, and run the graphical\
		environment using the command \"Hyprland\"."
}

### THE ACTUAL SCRIPT ###

### This is how everything happens in an intuitive format and order.

# Check if user is root on Arch distro. Install whiptail.
pacman --noconfirm --needed -Sy libnewt ||
	error "Are you sure you're running this as the root user, are on an Arch-based distribution and have an internet connection?"

# Welcome user and pick dotfiles.
welcomemsg || error "User exited."

# Get and verify username and password.
getuserandpass || error "User exited."

# Give warning if user already exists.
usercheck || error "User exited."

# Select package categories to install
selectcategories || error "User exited."

# Last chance for user to back out before install.
preinstallmsg || error "User exited."

### The rest of the script requires no user input.

# Refresh Arch keyrings.
refreshkeys ||
	error "Error automatically refreshing Arch keyring. Consider doing so manually."

whiptail --title "LARBS Installation" \
	--infobox "Installing software required to install and configure other programs." 8 70
pacman --noconfirm --needed -S curl ca-certificates base-devel git ntp

whiptail --title "LARBS Installation" \
	--infobox "Synchronizing system time to ensure successful and secure installation of software..." 8 70
ntpd -q -g >/dev/null 2>&1

adduserandpass || error "Error adding username and/or password."

[ -f /etc/sudoers.pacnew ] && cp /etc/sudoers.pacnew /etc/sudoers # Just in case

# Allow user to run sudo without password. Since AUR programs must be installed
# in a fakeroot environment, this is required for all builds with AUR.
trap 'rm -f /etc/sudoers.d/larbs-temp' HUP INT QUIT TERM PWR EXIT
echo "%wheel ALL=(ALL) NOPASSWD: ALL
Defaults:%wheel,root runcwd=*" >/etc/sudoers.d/larbs-temp

# Make pacman colorful, concurrent downloads and Pacman eye-candy.
grep -q "ILoveCandy" /etc/pacman.conf || sed -i "/#VerbosePkgLists/a ILoveCandy" /etc/pacman.conf
sed -Ei "s/^#(ParallelDownloads).*/\1 = 5/;/^#Color$/s/#//" /etc/pacman.conf

# Use all cores for compilation.
sed -i "s/-j2/-j$(nproc)/;/^#MAKEFLAGS/s/^#//" /etc/makepkg.conf

manualinstall $aurhelper || error "Failed to install AUR helper."

# Make sure .*-git AUR packages get updated automatically.
$aurhelper -Y --save --devel

# The command that does all the installing. Reads the progs.csv file and
# installs each needed program the way required. Be sure to run this only after
# the user has been created and has priviledges to run sudo without a password
# and all build dependencies are installed.
maininstallations

# Most important command! Get rid of the beep!
rmmod pcspkr
echo "blacklist pcspkr" >/etc/modprobe.d/nobeep.conf

# dbus UUID must be generated for Artix runit.
dbus-uuidgen >/var/lib/dbus/machine-id

# Allow wheel users to sudo with password and allow several system commands
# (like `shutdown` to run without password).
echo "%wheel ALL=(ALL:ALL) ALL" >/etc/sudoers.d/00-larbs-wheel-can-sudo
echo "%wheel ALL=(ALL:ALL) NOPASSWD: /usr/bin/shutdown,/usr/bin/reboot,/usr/bin/systemctl suspend,/usr/bin/wifi-menu,/usr/bin/mount,/usr/bin/umount,/usr/bin/pacman -Syu,/usr/bin/pacman -Syyu,/usr/bin/pacman -Syyu --noconfirm,/usr/bin/loadkeys,/usr/bin/pacman -Syyuw --noconfirm,/usr/bin/pacman -S -y --config /etc/pacman.conf --,/usr/bin/pacman -S -y -u --config /etc/pacman.conf --" >/etc/sudoers.d/01-larbs-cmds-without-password
echo "Defaults editor=/usr/bin/vim" >/etc/sudoers.d/02-larbs-visudo-editor
mkdir -p /etc/sysctl.d
echo "kernel.dmesg_restrict = 0" > /etc/sysctl.d/dmesg.conf

# Cleanup
rm -f /etc/sudoers.d/larbs-temp

# Last message! Install complete!
finalize
